/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <libetpan.h>

#include "x_alloc.h"
#include "logging.h"

#include "identity.h"
#include "mailbox.h"

typedef struct XePIdentityDisconnectData_s XePIdentityDisconnectData;

struct XePIdentityDisconnectData_s
{
  int count;
  XePIdentity *identity;
  XePIdentityDisconnectAllCb user_func;
  void *user_data;
};

/* Functions declarations */

static void
Identity_set_string(char **dest, const char *source);

static void
Identity_free(XePIdentity *identity, int copy);

static void
Identity_disconnect_all_cb(XePMailbox *mailbox, void *adata);

/* static global variables */

static clist *global_identities;

/* Function definitions */

static void
Identity_set_string(char **dest, const char *source)
{
  x_free(*dest);
  if (source)
    *dest = strdup(source);
  else
    *dest = NULL;
}

clist *
a_Identity_list_new(void)
{
  clist *identities;

  identities = clist_new();
  if (identities == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("clist_new failed");
    }
  global_identities = identities;

  return identities;
}

void
a_Identity_list_free(clist *identities)
{
  clistiter *iter;

  for (iter = clist_begin(identities); iter; iter = iter->next)
    {
      XePIdentity *identity = (XePIdentity *) iter->data;
      a_Identity_free(identity);
    }
  clist_free(identities);
}

clist *
a_Identity_list_get(void)
{
  return global_identities;
}

static void
Identity_signatures_free(clist *signatures)
{
  clistiter *iter;
  for (iter = clist_begin(signatures); iter; iter = iter->next)
    {
      x_free(iter->data);
    }
  clist_free(signatures);
}

XePIdentity *
a_Identity_new(void)
{
  XePIdentity *identity;

  identity = x_alloc(1, XePIdentity);
  if (identity == NULL)
    {
      A_LOGGING_SYSTEM_ERROR("x_alloc: %s", strerror(errno));
    }
  else
    {
      identity->id = NULL;
      identity->name = NULL;
      identity->email = NULL;
      identity->organization = NULL;
      identity->reply_to = NULL;
      identity->draft_ref = NULL;
      identity->queue_ref = NULL;
      identity->sent_ref = NULL;
      identity->mta_ref = NULL;
      identity->templates_ref = NULL;
      identity->signatures = clist_new();
      identity->mailboxes = a_Mailbox_list_new();
    }
  return identity;
}

XePIdentity *
a_Identity_new_copy(XePIdentity *identity)
{
  XePIdentity *identity_copy;
  clist *signatures_copy;
  clist *mailboxes_copy;

  A_LOGGING_INFO("Copying identity <%s>", identity->id);

  identity_copy = a_Identity_new();
  
  a_Identity_set_id(identity_copy, a_Identity_get_id(identity));
  a_Identity_set_name(identity_copy, a_Identity_get_name(identity));
  a_Identity_set_email(identity_copy, a_Identity_get_email(identity));
  a_Identity_set_organization(identity_copy, 
			      a_Identity_get_organization(identity));

  a_Identity_set_reply_to(identity_copy, a_Identity_get_reply_to(identity));
  a_Identity_set_draft_ref(identity_copy, a_Identity_get_draft_ref(identity));
  a_Identity_set_queue_ref(identity_copy, a_Identity_get_queue_ref(identity));
  a_Identity_set_sent_ref(identity_copy, a_Identity_get_sent_ref(identity));
  a_Identity_set_mta_ref(identity_copy, a_Identity_get_mta_ref(identity));

  signatures_copy = 
    a_Identity_signatures_new_copy(a_Identity_get_signatures(identity));

  a_Identity_set_signatures(identity_copy, signatures_copy);
			    
  mailboxes_copy = 
    a_Mailbox_list_new_copy(a_Identity_get_mailboxes(identity));

  a_Identity_set_mailboxes(identity_copy, mailboxes_copy);

  return identity_copy;
}

void
a_Identity_free(XePIdentity *identity)
{
  Identity_free(identity, 0);
}

void
a_Identity_free_copy(XePIdentity *identity)
{
  A_LOGGING_INFO("Freeing identity <%s>", identity->id);
  Identity_free(identity, 1);
}

static void
Identity_free(XePIdentity *identity, int copy)
{
  x_free(identity->id);
  x_free(identity->name);
  x_free(identity->email);
  x_free(identity->draft_ref);
  x_free(identity->queue_ref);
  x_free(identity->sent_ref);
  x_free(identity->mta_ref);

  Identity_signatures_free(identity->signatures);
  if (copy == 0)
    a_Mailbox_list_free(identity->mailboxes);
  else
    a_Mailbox_list_free_copy(identity->mailboxes);

  x_free(identity);
}

void
a_Identity_set_id(XePIdentity *identity, const char* id)
{
  Identity_set_string(&identity->id, id);
}

const char *
a_Identity_get_id(XePIdentity *identity)
{
  return identity->id;
}

void
a_Identity_set_name(XePIdentity *identity, const char* name)
{
  Identity_set_string(&identity->name, name);
}


const char *
a_Identity_get_name(XePIdentity *identity)
{
  return identity->name;
}

void
a_Identity_set_email(XePIdentity *identity, const char* email)
{
  Identity_set_string(&identity->email, email);
}

const char *
a_Identity_get_email(XePIdentity *identity)
{
  return identity->email;
}

void
a_Identity_set_organization(XePIdentity *identity, const char* organization)
{
  Identity_set_string(&identity->organization, organization);
}

const char *
a_Identity_get_organization(XePIdentity *identity)
{
  return identity->organization;
}

void
a_Identity_set_reply_to(XePIdentity *identity, const char* reply_to)
{
  Identity_set_string(&identity->reply_to, reply_to);
}

const char *
a_Identity_get_reply_to(XePIdentity *identity)
{
  return identity->reply_to;
}

void
a_Identity_set_draft_ref(XePIdentity *identity, const char* draft_ref)
{
  Identity_set_string(&identity->draft_ref, draft_ref);
}

const char *
a_Identity_get_draft_ref(XePIdentity *identity)
{
  return identity->draft_ref;
}

void
a_Identity_set_queue_ref(XePIdentity *identity, const char* queue_ref)
{
  Identity_set_string(&identity->queue_ref, queue_ref);
}

const char *
a_Identity_get_queue_ref(XePIdentity *identity)
{
  return identity->queue_ref;
}

void
a_Identity_set_sent_ref(XePIdentity *identity, const char* sent_ref)
{
  Identity_set_string(&identity->sent_ref, sent_ref);
}

const char *
a_Identity_get_sent_ref(XePIdentity *identity)
{
  return identity->sent_ref;
}

void
a_Identity_set_templates_ref(XePIdentity *identity, const char* templates_ref)
{
  Identity_set_string(&identity->templates_ref, templates_ref);
}

const char *
a_Identity_get_templates_ref(XePIdentity *identity)
{
  return identity->templates_ref;
}

void
a_Identity_set_mta_ref(XePIdentity *identity, const char *mta_ref)
{
  Identity_set_string(&identity->mta_ref, mta_ref);
}

const char *
a_Identity_get_mta_ref(XePIdentity *identity)
{
  return identity->mta_ref;
}

void
a_Identity_set_signatures(XePIdentity *identity, clist *signatures)
{
  Identity_signatures_free(identity->signatures);
  identity->signatures = signatures;
}

clist *
a_Identity_get_signatures(XePIdentity *identity)
{
  return identity->signatures;
}

clist *
a_Identity_signatures_new_copy(clist *signatures)
{
  clist *signatures_copy;
  clistiter *iter;
  char *sig;

  signatures_copy = clist_new();
  if (signatures_copy == NULL)
    {
      A_LOGGING_ERROR("malloc(): %s", strerror(errno));
      return NULL;
    }
  for (iter = clist_begin(signatures); iter; iter = iter->next)
    {
      if (iter->data)
	{
	  sig = strdup(iter->data);
	  if (sig)
	    {
	      if (clist_append(signatures_copy, sig))
		{
		  free(sig);
		  A_LOGGING_ERROR("malloc(): %s", strerror(errno));
		  Identity_signatures_free(signatures_copy);
		  return NULL;
		}
	    }
	  else
	    {
	      A_LOGGING_ERROR("strdup(): %s", strerror(errno));
	      Identity_signatures_free(signatures_copy);
	      return NULL;
	    }
	}
    }
  return signatures_copy;
}

void
a_Identity_set_mailboxes(XePIdentity *identity, clist *mailboxes)
{
  a_Mailbox_list_free(identity->mailboxes);
  identity->mailboxes = mailboxes;
}

clist *
a_Identity_get_mailboxes(XePIdentity *identity)
{
  return identity->mailboxes;
}

XePIdentity *
a_Identity_lookup_by_email(clist *identities, const char *email)
{
  struct mailimf_mailbox *mb = NULL;
  size_t cur_token = 0;
  clistiter *iter = NULL;
  XePIdentity *identity = NULL;

  if (identities == NULL)
    return NULL;

  if ((email == NULL) || 
      (MAILIMF_NO_ERROR == mailimf_mailbox_parse((char *) email, 
						 strlen(email), 
						 &cur_token,
						 &mb)))
    {
      for (iter = clist_begin(identities); iter; iter = iter->next)
	{
	  identity = (XePIdentity *) iter->data;
	  if (((mb->addr_spec)   &&
	       (identity->email) && 
	       (strcmp(mb->addr_spec, identity->email) == 0)) ||
	      (email == NULL))
	    {
	      break;
	    }
	}
      mailimf_mailbox_free(mb);
    }

  return (iter ? identity : NULL);
}

int
a_Identity_is_busy(XePIdentity *identity)
{
  clistiter *iter;
  XePMailbox *mailbox;

  for (iter = clist_begin(identity->mailboxes); iter; iter = iter->next)
    {
      mailbox = (XePMailbox *) iter->data;
      if (a_Mailbox_is_busy(mailbox))
	return -1;
    }
  return 0;
}

void
a_Identity_disconnect_all(XePIdentity *identity, 
			  XePIdentityDisconnectAllCb user_func,
			  void *user_data)
{
  clistiter *iter;
  XePMailbox *mailbox;
  XePIdentityDisconnectData *data;
  int count = clist_count(identity->mailboxes);

  if (count == 0)
    {
      if (user_func)
	user_func(identity, user_data);
      return;
    }

  data = x_alloc(1, XePIdentityDisconnectData);
  data->count = count;
  data->identity = identity;
  data->user_func = user_func;
  data->user_data = user_data;

  for (iter = clist_begin(identity->mailboxes); iter; iter = iter->next)
    {
      mailbox = (XePMailbox *) iter->data;
      a_Mailbox_disconnect_all(mailbox, Identity_disconnect_all_cb, data);
    }
}

static void
Identity_disconnect_all_cb(XePMailbox *mailbox, void *adata)
{
  XePIdentityDisconnectData *data = (XePIdentityDisconnectData *) adata;

  data->count--;
  if (data->count)
    return; 			/* Still mailboxes connected */

  /* All disconnected */

  if (data->user_func)
    data->user_func(data->identity, data->user_data);

  x_free(data);
}
