/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef PROCESSING_H
#define PROCESSING_H

#include <chash.h>

#include "cntree.h"

/*! Main data structure containing all processing info */
typedef struct XePProcessing_s XePProcessing;
struct XePProcessing_s
{
  chash *rules;        /**< Processing Rules.      */
  chash *conditions;   /**< Processing Conditions. */
  chash *actions;      /**< Processing Actions.    */
};

/*! Processing condition. 
 *  The conditions are used for filtering and message management.
 *  They are also used for interactive search and by virtual folders.
 */
typedef struct XePProcessingCondition_s XePProcessingCondition;
struct XePProcessingCondition_s
{
  char *id;          /**< Name of the condition.  */
  char *expression;  /**< The condition string.  */
  int enabled;       /**< Global "enabled" flag.  */
};

/*! Processing action. 
 *  The action is executed when a condition is met, but also
 *  when the user explicitely ask it or program it..
 */
typedef struct XePProcessingAction_s XePProcessingAction;
struct XePProcessingAction_s
{
  char *id;      /**< The name of the action.  */
  char *command; /**< The command to be executed.  */
  int enabled;   /**< Global "enabled" status.  */
};

/*! Processing rule.
 *  A rule is the combination of a condition and an action.
 *  The action is run only when the condition is met.
 */
typedef struct XePProcessingRule_s XePProcessingRule;
struct XePProcessingRule_s
{
  char *id;                          /**< The name of the rule.  */
  XePProcessingCondition *condition; /**< Pointer to a condition.  */
  XePProcessingAction *action;       /**< Pointer to an action.  */
  char *condition_ref;               /**< Reference to a
				      *   condition. Ignored if the
				      *   condition member is not
				      *   NULL.  */
  char *action_ref;                  /**< Reference to an
				      *   action. Ignored if the
				      *   action member is not
				      *   NULL.  */
  int enabled;                       /**< Global "enabled" flag.  */
};

/*< Used to reference a rule (real folder) or a condition (virtual folder) */
typedef struct XePProcessingRef_s XePProcessingRef;
struct XePProcessingRef_s
{
  char *ref;
  int enabled; /**< "Enabled" flag. Default is yes (1).  */
  int periodicity;
};


XePProcessing *
a_Processing_new(void);

XePProcessingCondition *
a_Processing_condition_new(void);

XePProcessingAction *
a_Processing_action_new(void);

XePProcessingRule *
a_Processing_rule_new(void);

XePProcessingRef *
a_Processing_ref_new(void);

XePProcessingRef *
a_Processing_ref_new_copy(XePProcessingRef *ref);

/*! Release processing resources. */
void
a_Processing_free(XePProcessing *processing);

/*! Release condition resources. */
void
a_Processing_condition_free(XePProcessingCondition *condition);

/*! Release action resources. */
void
a_Processing_action_free(XePProcessingAction *action);

/*! Release rule resources. */
void
a_Processing_rule_free(XePProcessingRule *action);

void
a_Processing_ref_free(XePProcessingRef *ref);

clist *
a_Processing_ref_list_new(void);

void
a_Processing_ref_list_free(clist *refs);

int
a_Processing_ref_list_copy(clist *source, clist *target);

/*! Add a condition to the processing list */
int
a_Processing_condition_add(XePProcessing *processing,
			   XePProcessingCondition *condition);

/*! Add an action to the processing list */
int
a_Processing_action_add(XePProcessing *processing,
			XePProcessingAction *action);
/*! Add a rule to the processing list */
int
a_Processing_rule_add(XePProcessing *processing,
		      XePProcessingRule *rule);

/*! Create processing data from XML N-tree. */
int
a_Processing_create(cntree *ntree, XePProcessing **processing, int op);

int
a_Processing_update_rules_links(XePProcessing *processing);

XePProcessingRule *
a_Processing_rule_get_from_ref(chash *rules, char *ref);

XePProcessingCondition *
a_Processing_condition_get_from_ref(chash *conditions, char *ref);

#endif
