/*
 * XetPan: a libEtPan! based Mail User Agent
 * Copyright (C) 2002-2003 Melvin Hadasht
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h> /* open */
#include <sys/stat.h>  /* open, fstat */
#include <fcntl.h>     /* open*/

#include <sys/wait.h>  /* waitpid */
#include <unistd.h> /* read, write, close */

#ifdef HAVE_MAGIC_H
#include <magic.h>
/* There is a bug in magic_buffer: it has an internal buffer size of
 * 65k but it does not check for buffer overflow. Limit it to 10000.
 */
#define MAGIC_MAX_BUF_LEN 10000
#endif

#include <clist.h>

#include "defines.h"

#include "intl.h"
#include "logging.h"
#include "x_alloc.h"
#include "utils.h"

/*! Max line length supported by SMTP servers, not including \r\n */
#define SMTP_MAX_LINE_LEN 998

/* The "begin" line length of uuencoded or yEncoded data should not
 * exceed MAX_ALLOWED_BEGIN_LINE_LEN. An empirical value of 1023
 * is chosen. We use 1023 instead of 1024 because a char buffer is 
 * locally created, and it should contain the trailing zero.
 * (Theoretically it should be less than SMTP_MAX_LINE_LEN)
 */
#define MAX_ALLOWED_BEGIN_LINE_LEN (1024 - 1)

extern long int timezone;

/* Known charset encodings. The order must match the 
 * ones in XePUtilsCharset enumeration
 */
XePUtilsCharsetMapping xetpan_charsets[] = 
  {
    {XEP_CHARSET_0, NULL},
    {XEP_CHARSET_BIG5, "BIG5"},
    {XEP_CHARSET_EUC_JP,"EUC-JP"},
    {XEP_CHARSET_EUC_KR,"EUC-KR"},
    {XEP_CHARSET_EUC_TW,"EUC-TW"},
    {XEP_CHARSET_GB2312,"GB2312"},
    {XEP_CHARSET_ISIRI_3342,"ISIRI-3342"},
    {XEP_CHARSET_ISO_2022_CN, "ISO-2022-CN"},
    {XEP_CHARSET_ISO_2022_CN_EXT, "ISO-2022-CN-EXT"},
    {XEP_CHARSET_ISO_2022_JP, "ISO-2022-JP"},
    {XEP_CHARSET_ISO_2022_JP_2, "ISO-2022-JP-2"},
    {XEP_CHARSET_ISO_2022_KR, "ISO-2022-KR"},
    {XEP_CHARSET_ISO_8859_1, "ISO-8859-1"},
    {XEP_CHARSET_ISO_8859_10, "ISO-8859-10"},
    {XEP_CHARSET_ISO_8859_11, "ISO-8859-11"},
    {XEP_CHARSET_ISO_8859_13, "ISO-8859-13"},
    {XEP_CHARSET_ISO_8859_14, "ISO-8859-14"},
    {XEP_CHARSET_ISO_8859_15, "ISO-8859-15"},
    {XEP_CHARSET_ISO_8859_16, "ISO-8859-16"},
    {XEP_CHARSET_ISO_8859_2, "ISO-8859-2"},
    {XEP_CHARSET_ISO_8859_3, "ISO-8859-3"},
    {XEP_CHARSET_ISO_8859_4, "ISO-8859-4"},
    {XEP_CHARSET_ISO_8859_5, "ISO-8859-5"},
    {XEP_CHARSET_ISO_8859_6, "ISO-8859-6"},
    {XEP_CHARSET_ISO_8859_7, "ISO-8859-7"},
    {XEP_CHARSET_ISO_8859_8, "ISO-8859-8"},
    {XEP_CHARSET_ISO_8859_8_I, "ISO-8859-I"},
    {XEP_CHARSET_ISO_8859_9, "ISO-8859-9"},
    {XEP_CHARSET_KOI8_R, "KOI8-R"},
    {XEP_CHARSET_KOI8_U, "KOI8-U"},
    {XEP_CHARSET_MAC_CYRILLIC, "MAC-CYRILLIC"},
    {XEP_CHARSET_SHIFT_JIS, "SHIFT-JIS"},
    {XEP_CHARSET_TIS_620, "TIS-620"},
    {XEP_CHARSET_UTF8, "UTF-8"},
    {XEP_CHARSET_VISCII, "VISCII"}, 
    {XEP_CHARSET_WINDOWS_1250, "WINDOWS-1250"},
    {XEP_CHARSET_WINDOWS_1251, "WINDOWS-1251"},
    {XEP_CHARSET_WINDOWS_1254, "WINDOWS-1254"},
    {XEP_CHARSET_WINDOWS_1255, "WINDOWS-1255"},
    {XEP_CHARSET_WINDOWS_1256, "WINDOWS-1256"},
    {XEP_CHARSET_WINDOWS_1258, "WINDOWS-1258"},
    {XEP_CHARSET_0, NULL}
  };

typedef XePUtilsInlinedData * (*XePUtilsTextScanFunc)(const char *data, 
						    size_t size);
/* Static functions declarations */

static XePUtilsInlinedData *
Utils_data_has_yencoded(const char *data,
			size_t size,
			XePUtilsInlinedDataType type);

static int
Utils_is_CTL(unsigned char c);

static int
Utils_is_tspecials(unsigned char c);

static int
Utils_is_SPACE(unsigned char c);

static int
Utils_is_WSP(unsigned char c);

static int
Utils_is_NO_WS_CTL(unsigned char c);

static int
Utils_is_ALPHA(unsigned char c);

static int
Utils_is_DIGIT(unsigned char c);

static int
Utils_is_quoted_pair(unsigned char *s);

static int
Utils_is_text(unsigned char c);

static int
Utils_is_atext(unsigned char c);

static int
Utils_is_ctext(unsigned char c);

static int
Utils_is_qtext(unsigned char c);

static int
Utils_is_qcontent(unsigned char *s);

static int
Utils_is_utext(unsigned char c);

static int
Utils_is_specials(unsigned char c);

static int
Utils_is_especials(unsigned char c);

static int
Utils_is_rfc822_quoted_pair(unsigned char *s);

static int
Utils_is_rfc822_text(unsigned char *s);

static int
Utils_is_to_be_q_encoded(char c, int is_phrase, int is_comment);

static XePUtilsEncodingType
Utils_select_b_q_encoding(const char *text, const char *charset, 
			  int is_phrase, int is_comment);

/* Function definitions */

char *
a_Utils_printf(char *template, XePUtilsFormat *format)
{
  char *text;
  char *text_p;
  char *template_p;
  XePUtilsFormat *a_format;
  int len;

  len = strlen(template);

  template_p = template;

  while (*template_p && (template_p = strchr(template_p, '%')))
    {
      template_p++;
      if (*template_p && *template_p != '%')
	{
	  for (a_format = format; a_format->token; a_format++)
	    {
	      if (a_format->token == *template_p)
		{
		  len = len - 2 + (a_format->string ? 
				   strlen(a_format->string) :
				   0);
		  break;
		}
	    }
	  template_p++;
	}
    }

  text = (char *) malloc(len + 1);

  template_p = template;
  text_p = text;
 
  while (*template_p)
    {
      *text_p = *template_p;

      if (*template_p == '%')
	{
	  template_p++;
	  if (*template_p)
	    {
	      for (a_format = format; a_format->token; a_format++)
		{
		  if (a_format->token == *template_p)
		    {
		      if (a_format->string)
			{
			  *text_p = 0x00;
			  strcat(text_p, a_format->string);
			  text_p += strlen(a_format->string);
			}
		      template_p++;
		      break;
		    }
		}
	      if (!a_format->token)
		{
		  text_p++;
		  *text_p = *template_p;
		  template_p++;
		  text_p++;
		}
	    } 
	}
      else
	{
	  text_p++;
	  template_p++;
	}
    }
  *text_p = 0x00;
  return text;
}

char *
a_Utils_append(char *dest, ...)
{
  va_list args;
  char *chunk;
  int len = 0; 
  char *result;

  if (dest)
    len = strlen(dest);

  va_start(args, dest);

  while ((chunk = va_arg(args, char *)))
    {
      len += strlen(chunk);
    }
  va_end(args);

  /* TODO: Error check */
  if (dest)
    {
      result = realloc(dest, len + 1);
    }
  else
    {
      result = x_alloc(len + 1, char);
      result[0] = 0x00;
    }

  va_start(args, dest);
  while ((chunk = va_arg(args, char *)))
    {
      strcat(result, chunk);
    }
  va_end(args);
  return result;
}

/*! FIXME: this function should be called 'clear_crlf' */
void
a_Utils_string_unfold(char *string)
{
  char *stripped;
  char *original;
  int handled = 0;
  
  for (stripped = string, original = string; 
       *stripped && *original; 
       original++)
    {
      if ((*original == '\n') || (*original == '\r') || (*original == '\t'))
	{
	  original++;
	  if (!handled)
	    {
	      *stripped++ = ' ';
	      handled = 1;
	    }
	}
      else
	{
	  *stripped++ = *original;
	  handled = 0;
	}
    }
  *stripped = 0x00;
}

char *
a_Utils_magic_data_mime_type(const char *data, size_t size)
{
  char *mime_type;

#ifdef HAVE_MAGIC_H
  magic_t magic;
  const char *magic_type;

  magic = magic_open(MAGIC_MIME);

  magic_load(magic, NULL);
  magic_type = magic_buffer(magic, data,
			    ((size < MAGIC_MAX_BUF_LEN) ?
			     size :
			     MAGIC_MAX_BUF_LEN));

  if (magic_type)
    mime_type = strdup(magic_type);
  else
    mime_type= strdup("application/octet-stream");

  magic_close(magic);

#else 
  /* TODO: if no libmagic is present, use another way to guess the MIME type */
  mime_type= strdup("application/octet-stream");

#endif
  return mime_type;
}

char *
a_Utils_magic_file_mime_type(const char *filename)
{
  char *mime_type;

#ifdef HAVE_MAGIC_H
  magic_t magic;
  const char *magic_type;

  magic = magic_open(MAGIC_MIME);

  magic_load(magic, NULL);
  magic_type = magic_file(magic, filename);

  if (magic_type)
    mime_type = strdup(magic_type);
  else
    mime_type= strdup("application/octet-stream");

  magic_close(magic);

#else
  /* TODO: if no libmagic is present, use another way to guess the MIME type */
  mime_type= strdup("application/octet-stream");
#endif

  return mime_type;
}

char *
a_Utils_magic_get_charset(const char *magic_mime_type)
{
  char *p;
  char *charset = NULL;
  const char *charset_token = "charset=";

  if (magic_mime_type == NULL)
    return NULL;

  p = strstr(magic_mime_type, charset_token);
  
  /* FIXME: charset value can be quoted and can contain commenrs
   * in accordance with RFC 822. The comments should be filtered
   * out.
   */
  if (p)
    {
      int i = 0;
      p += strlen(charset_token);
      for (i = 0; p[i] && p[i] != ';'; i++); /* Emtpy for block */
      if (i)
	{
	  charset = x_alloc(i + 1, char);
	  memcpy(charset, p, i);
	  charset[i] = 0x00;
	}
    }
  return charset;
}

void
a_Utils_format_size(char *buf, size_t size)
{
  
  /* b: byte
   * k: kilo byte = 1024 ~ 1e3
   * M: Mega byte = 1024*1024 ~ 1e6
   * G: Giga byte = 1024*1024*1024 ~ 1e9
   * P: Peta byte = 1024*1024*1024*1024 ~ 1e12
   * E: Exa byte  = 1024*1024*1024*1024*1024 ~ 1e15
   * Z: Zetta byte= 1024*1024*1024*1024*1024*1024 ~ 1e18
   * Y: Yotta byte= 1024*1024*1024*1024*1024*1024*1024 ~ 1e21 !
   */
  char *unit = "bkMGTPEZY?";
  
  for (; *unit && size > (1 << 10); unit++)
    {
      size = size >> 10;
    }
  sprintf(buf, "%4d%c", size, *unit);
}

int
a_Utils_write(int fd, const char *data, size_t size)
{
  size_t count = 0;
  size_t total_written = 0;
  int ok = 1;

  while (total_written != size && ok)
    {
      count = write(fd, data + total_written, size - count);
      if (count >= 0)
	{
	  total_written += count;
	}
      else
	{
	  A_LOGGING_SYSTEM_ERROR("write: %s", strerror(errno));
	  ok = 0;
	}
    }
  /* Return 0 on success, -1 on error */
  return (ok == 0);
}

char *
a_Utils_save_temp(const char *data, size_t size)
{
  int fd;
  char template[] = "xetpan.temp.XXXXXX";
  char *tempfile;
  int r;

  if (!data || size == 0)
    return NULL;

  fd = mkstemp(template);

  if (fd == -1)
    {
      A_LOGGING_SYSTEM_ERROR("mkstemp: %s", strerror(errno));
      return NULL;
    }

  r = a_Utils_write(fd, data, size);

  close(fd);

  if (r)
    tempfile = NULL;
  else
    tempfile = strdup(template);

  return tempfile;
}

int
a_Utils_read(int fd, char **data, size_t *size)
{
  struct stat stat_buf;
  char *buf;
  int r = -1;

  if (fstat(fd, &stat_buf) == 0)
    {
      if (stat_buf.st_size)
	{
	  size_t count;
	  size_t total = 0;
	  buf = x_alloc(stat_buf.st_size, char);
	  if (buf)
	    {
	      while ((total < stat_buf.st_size) &&
		     (0 <= (count = read(fd, buf + total, 
					 stat_buf.st_size - total))))
		{
		  total += count;
		}
	      if (count >= 0)
		{
		  *data = buf;
		  *size = stat_buf.st_size;
		  r = 0;
		}
	      else
		{
		  A_LOGGING_ERROR(_("Error while reading file"));
		  free(buf);
		}
	    }
	  else
	    {
	      A_LOGGING_SYSTEM_ERROR("x_alloc(): %s", strerror(errno));
	    }
	}
      else
	{
	  *data = NULL;
	  *size = 0;
	  r = 0;
	}
    }
  return r;
}
  
int
a_Utils_load_file(const char *filename, char **data, size_t *size)
{
  int fd;
  int r = -1;
  
  fd = open(filename, O_RDONLY);
  if (fd)
    {
      r = a_Utils_read(fd, data, size);
      if (r)
	{
	  A_LOGGING_ERROR(_("File '%s' could not be read."), filename);
	}
      close(fd);
    }
  else
    {
      A_LOGGING_ERROR(_("Error while opening file '%s': %s"), filename,
		      strerror(errno));
    }

  return r;  
}

int
a_Utils_file_backup(const char *filename)
{
  int r = -1;
  char *new_name;
  const char *backup_suffix = ".bak";
  struct stat file_stat;

  if (stat(filename, &file_stat))
    {
      if (errno == ENOENT)
	return 0;
      A_LOGGING_ERROR("file '%s': %s",
		      filename ? filename : "<EMTPY FILENAME>",
		      strerror(errno));
      return -1;
    }
		      
  
  new_name = x_alloc(strlen(filename) + strlen(backup_suffix) + 1, char);

  if (new_name)
    {
      strcpy(new_name, filename);
      strcat(new_name, backup_suffix);
      r = rename(filename, new_name);
      free(new_name);
      if (r)
	{
	  A_LOGGING_ERROR("Error while renaming '%s': rename() failed: %s",
			  filename, strerror(errno));
	}
    }
  else
    {
      A_LOGGING_ERROR("x_alloc() failed: %s", strerror(errno));
    }
  return r;
}

XePUtilsInlinedData *
a_Utils_inlined_text_new(XePUtilsInlinedDataType type,
			 const char *data,
			 size_t size,
			 const char *plain_text,
			 size_t plain_size,
			 int privacy_sig_status,
			 int privacy_enc_status,
			 void *extra)
{
  XePUtilsInlinedText *inlined_data;

  if (type != XEP_INLINED_DATA_TEXT &&
      type != XEP_INLINED_DATA_CLEARSIGNED_TEXT &&
      type != XEP_INLINED_DATA_ASCII_ARMORED_TEXT)
    {
      A_LOGGING_ERROR(_("Incorrect XEP_INLINED_DATA_* type for inlined text"));
      return NULL;
    }

  if (data == NULL || *data == 0x00 || size == 0)
    return NULL;

  inlined_data = x_alloc(1, XePUtilsInlinedText);
  if (inlined_data)
    {
      inlined_data->type = type;
      inlined_data->extra = extra;
      inlined_data->start = data;
      inlined_data->privacy_sig_status = privacy_sig_status;
      inlined_data->privacy_enc_status = privacy_enc_status;
      inlined_data->plain_size = size;

      inlined_data->data = x_alloc(size, char);
      if (inlined_data->data)
	{
	  memcpy(inlined_data->data, data, size);
	  inlined_data->size = size;
	  if (plain_size)
	    {
	      inlined_data->plain_text = x_alloc(plain_size, char);
	      if (inlined_data->plain_text)
		{
		  memcpy(inlined_data->plain_text, plain_text,
			 plain_size);
		}
	      else
		{
		  x_free(inlined_data->data);
		  x_free(inlined_data);
		  inlined_data = NULL;
		}
	    }
	  else
	    {
	      inlined_data->plain_text = NULL;
	    }
	}
      else
	{
	  x_free(inlined_data->data);
	  x_free(inlined_data);
	  inlined_data = NULL;
	}
    }
  else
    {
      A_LOGGING_ERROR("x_alloc(): %s", strerror(errno));
    }
      
  return (XePUtilsInlinedData *) inlined_data;
}

XePUtilsInlinedData *
a_Utils_inlined_file_new(XePUtilsInlinedDataType type,
			 const char *data,
			 size_t size,
			 const char *filename,
			 void *extra)
{
  XePUtilsInlinedFile *inlined_data;

  if (filename == NULL || data == NULL || size == 0)
    return NULL;

  if (type != XEP_INLINED_DATA_UUENCODED &&
      type != XEP_INLINED_DATA_YENCODED &&
      type != XEP_INLINED_DATA_YENCODED2 &&
      type != XEP_INLINED_DATA_BINHEX)
    {
      A_LOGGING_ERROR(_("Incorrect XEP_INLINED_DATA_* type for inlined file"));
      return NULL;
    }

  inlined_data = x_alloc(1, XePUtilsInlinedFile);
  if (inlined_data)
    {
      inlined_data->filename = strdup(filename);
      inlined_data->type = type;
      inlined_data->start = data;
      inlined_data->data = x_alloc(size, char);
      if (inlined_data->data)
	{
	  memcpy(inlined_data->data, data, size);
	  inlined_data->size = size;
	  inlined_data->extra = extra;
	}
      else
	{
	  x_free(inlined_data->filename);
	  x_free(inlined_data);
	  inlined_data = NULL;
	}
    }
  else
    {
      A_LOGGING_ERROR("x_alloc(): %s", strerror(errno));
    }

  return (XePUtilsInlinedData *) inlined_data;
}

void
a_Utils_inlined_list_free(clist *inlined_list)
{
  clist_foreach(inlined_list, (clist_func) a_Utils_inlined_data_free, NULL);
  clist_free(inlined_list);
}

void
a_Utils_inlined_data_free(XePUtilsInlinedData *inlined_data)
{
  if (inlined_data == NULL)
    return;

  x_free(inlined_data->any.data);

  switch (inlined_data->any.type)
    {
    case XEP_INLINED_DATA_TEXT:
    case XEP_INLINED_DATA_CLEARSIGNED_TEXT:
    case XEP_INLINED_DATA_ASCII_ARMORED_TEXT:
      x_free(inlined_data->text.plain_text);
      break;
    case XEP_INLINED_DATA_UUENCODED:
    case XEP_INLINED_DATA_YENCODED:
    case XEP_INLINED_DATA_YENCODED2:
    case XEP_INLINED_DATA_BINHEX:
      x_free(inlined_data->file.filename);
      break;
    default:
      A_LOGGING_SYSTEM_ERROR("Internal error: attempt to free unknown inlined "
			     "data");
    }
  x_free(inlined_data);
}

int
a_Utils_inlined_file_decode_as(XePUtilsInlinedData *inlined_data,
			       const char *save_as)
{
  char *tempfile;
  char *filename;
  int r = 0;

  if (inlined_data == NULL)
    return r;

  tempfile = a_Utils_save_temp(inlined_data->any.data, inlined_data->any.size);
  if (tempfile)
    {
      filename = a_Utils_uulib_decode_file(tempfile, NULL, NULL, NULL);
      unlink(tempfile);
      x_free(tempfile);
      if (filename)
	{
	  if (save_as)
	    {
	      if (rename(filename, save_as))
		{
		  A_LOGGING_ERROR(_("Error while renaming file '%s' to '%s': %s"),
				  filename, save_as, strerror(errno));
		  r = -1;

		}
	    }
	  x_free(filename);
	}
      else
	{
	  r = -1;
	}
    }
  else
    {
      r = -1;
    }

  return r;
}

char *
a_Utils_uulib_decode_file(const char *filename,
			  void (*msg_cb)(void *adata, 
					 const char *msg,
					 int msg_type),
			  void (*busy_cb)(void *adata,
					  uuprogress *aprogress),
			  void *data)
{
  int r;
  char *decoded_file = NULL;
  uulist *item;


  UUInitialize();


  if (msg_cb)
    UUSetMsgCallback(data, msg_cb);

  if (busy_cb)
    UUSetBusyCallback(data, busy_cb);

  r = UULoadFile(filename, NULL, 0);
  if (r != UURET_OK)
    {
      A_LOGGING_ERROR("Could not load file %s to decode: %s",
		      filename, 
		      (r == UURET_IOERR ? 
		       strerror(UUGetOption (UUOPT_ERRNO, NULL, NULL, 0)):
		       UUstrerror(r)));
      UUCleanUp();
      return NULL;
    }

  item = UUGetFileListItem(0);
  if (item && (item->state & UUFILE_OK) != 0)
    {
      r = UUDecodeFile(item, NULL);
      if (r != UURET_OK)
	{
	  A_LOGGING_ERROR("Error while decoding file %s: %s",
			  (item->filename == NULL ?
			   "<File with no name>" :
			   item->filename),
			  (r == UURET_IOERR ? 
			   strerror(UUGetOption (UUOPT_ERRNO, NULL, NULL, 0)):
			   UUstrerror(r)));
	  UUCleanUp();
	  return NULL;
	}
      else
	{
	  decoded_file = strdup(item->filename);
	}
    }
  else
    {
      A_LOGGING_ERROR(_("Error when attempting to decode file %s: %s"),
			filename,
			strerror(UUGetOption (UUOPT_ERRNO, NULL, NULL, 0)));
    }
		      
			

  UUCleanUp();

  return decoded_file;
}

XePUtilsInlinedData *
a_Utils_data_has_uuencoded(const char *data, size_t size)
{
  const char *data_end = data + size;
  const char *begin_tag = "begin ";
  const char *end_tag = "end";
  const char *begin;
  const char *begin_eol;
  const char *mode;
  const char *end;
  const char *filename;
  int filename_len;
  char *inlined_filename;
  size_t inlined_size;
  XePUtilsInlinedData *result;
  int i;
  int ok = -1;

  if (data == NULL || size == 0)
    return NULL;
  
  begin = a_Utils_memmem(data, size, begin_tag, strlen(begin_tag));
  if (!begin)
    return NULL;

  /* Fail if "begin " is not at start of line */

  if (! (begin == data || *(begin - 1) == '\n' || *(begin - 1) == '\r'))
    return NULL;
  

  /* Bracket begin line between begin and begin_eol */

  mode = begin + strlen(begin_tag);

  begin_eol = mode;
  while ((begin_eol < data_end) && (*begin_eol != '\n' && *begin_eol != '\r'))
    {
      begin_eol++;
    }

  /* Fail if "begin " is on the last line of data */

  if (begin_eol == data_end)
    {
      return NULL;
    }
  
  /* Check for correct mode syntax */

  for (i = 0; i < 4 && mode[i] && mode + i < data_end && ok; i++)
    {
      if ((mode[i] < '0' || mode[i] > '7') &&
	  !(i == 3 && mode[i] == ' '))
	ok = 0;
    }
  if (!ok || mode + i == data_end)
    {
      return NULL;
    }

  /* Search for the filename */

  if (mode[3] == ' ')
    filename = mode + 3;
  else
    filename = mode + 4;

  while (*filename == ' ')
    filename++;
  
  /* Fail if there is no filename */

  if (filename == begin_eol)
    return NULL;
  
  filename_len = begin_eol - filename;
  
  inlined_filename = x_alloc(filename_len + 1, char);
  strncpy(inlined_filename, filename, filename_len);
  inlined_filename[filename_len] = 0x00;

  /* Find "end" tag */

  end = begin_eol;
  while((end = a_Utils_memmem(end, data + size - end, 
			      end_tag, strlen(end_tag))) && 
	(*(end - 1) != '\n' && *(end - 1) != '\r'))
    {
      end++;
    }
  
  if (end)
    {
      /* Size of data from "begin" to "end" */
      inlined_size = end + strlen(end_tag) - begin;
    }
  else
    {
      /* Size is the remaining size from "begin" */
      inlined_size = size - (begin - data);
    }

  result = a_Utils_inlined_file_new(XEP_INLINED_DATA_UUENCODED,
				    begin, inlined_size,
				    inlined_filename, NULL);
  x_free(inlined_filename);
  return result;
}

XePUtilsInlinedData *
a_Utils_data_has_yencoded(const char *data, size_t size)
{
  return Utils_data_has_yencoded(data, size, XEP_INLINED_DATA_YENCODED);
}

XePUtilsInlinedData *
a_Utils_data_has_yencoded2(const char *data, size_t size)
{
  return Utils_data_has_yencoded(data, size, XEP_INLINED_DATA_YENCODED2);
}

/* type is either XEP_INLINED_DATA_YENCODED or XEP_INLINED_DATA_YENCODED2 */
static XePUtilsInlinedData *
Utils_data_has_yencoded(const char *data, 
			size_t size,
			XePUtilsInlinedDataType type)
{
  const char *data_end = data + size;
  const char *begin_tag1 = "=ybegin ";
  const char *begin_tag2 = "=ybegin2 ";
  const char *begin_tag;
  const char *begin;
  const char *begin_eol;
  const char *end_tag = "=yend ";
  const char *end;
  const char *end_eol;
  const char *filename;
  int filename_len;
  char buffer[MAX_ALLOWED_BEGIN_LINE_LEN + 1];
  const char *size_tag;
  const char *line_tag;
  const char *name_tag;
  char *inlined_filename;
  XePUtilsInlinedData *result;

  if (data == NULL || size == 0)
    return NULL;

  begin_tag = (type == XEP_INLINED_DATA_YENCODED ? begin_tag1 : begin_tag2);

  /* We should loop until no begin is found */  
  begin = a_Utils_memmem(data, size, begin_tag, strlen(begin_tag));
  if (!begin)
    return NULL;

  /* Fail if "=ybegin " is not at start of line */
  if (! (begin == data || *(begin - 1) == '\n' || *(begin - 1) == '\r'))
    return NULL;
  

  /* Bracket begin line between begin and begin_eol */

  begin_eol = begin + strlen(begin_tag);

  while (begin_eol < data_end && (*begin_eol != '\n' && *begin_eol != '\r'))
    {
      begin_eol++;
    }
  /* Fail if "begin " is on the last line of data */
  if (begin_eol == data_end)
    {
      return NULL;
    }
  
  /* Fail if the begin line is too long: something is wrong or it is not
   * a yEnc block
   */
  if (begin_eol - begin > MAX_ALLOWED_BEGIN_LINE_LEN)
    {
      return NULL;
    }

  /* Check loosely for correct ybegin syntax by checking for 
   * "line=", "size=" and "name=".  The latter must be the last one 
   */
  memcpy(buffer, begin, begin_eol - begin);
  buffer[begin_eol - begin] = 0x00;
  size_tag = strstr(buffer, "size=");
  line_tag = strstr(buffer, "line=");
  name_tag = strstr(buffer, "name=");

  /* All three tags must be present */

  if (!size_tag || !line_tag || !name_tag)
    {
      return NULL;
    }

  /* The name tag must be the last tag */

  if (name_tag < size_tag || name_tag < line_tag)
    {
      return NULL;
    }

  /* Look for filename and fail if there is no filename */

  filename = begin + (name_tag - buffer) + strlen("name=");

  if (filename == begin_eol)
    {
      return NULL;
    }

  /* Search for the filename */

  while (*filename == ' ')
    filename++;
  
  /* Fail if there is no filename */

  if (filename == begin_eol)
    return NULL;
  
  filename_len = begin_eol - filename;

  /* Find "=yend " tag */

  end = begin_eol;
  while((end = a_Utils_memmem(end, data + size - end,
			      end_tag, strlen(end_tag))) && 
	(*(end - 1) != '\n' && *(end - 1) != '\r'))
    {
      end++;
    }

  /* end tag is mandatory in yEnc */

  if (!end)
    {
      return NULL;
    }

  end_eol = end + strlen(end_tag);
  while (end_eol < data_end && (*end_eol != '\n' && *end_eol != '\r'))
    {
      end_eol++;
    }

  /* Fail if "=yend " does not have a trailing CR/LF */

#ifdef STRICT_YENC
    if (end_eol == data_end)
    {
      return NULL;
    }
#endif
  
  /* Fail if the end line is too long: something is wrong or it is not
   * a yEnc block
   */
  if (end_eol - end > MAX_ALLOWED_BEGIN_LINE_LEN)
    {
      return NULL;
    }
  
  /* Check loosely for correct yend syntax by checking for 
   * "size="
   */
  memcpy(buffer, end, end_eol - end);
  buffer[end_eol - end] = 0x00;
  size_tag = strstr(buffer, "size=");

  /* "size=" tag is required */
  if (!size_tag)
    {
      return NULL;
    }

  /* Everything went OK so far. Retrieve filename */
  /* TODO: should we remove spaces and quotes? */

  inlined_filename = x_alloc(filename_len + 1, char);
  strncpy(inlined_filename, filename, filename_len);
  inlined_filename[filename_len] = 0x00;

  result = a_Utils_inlined_file_new(type, begin, end_eol - begin,
				    inlined_filename, NULL);
  x_free(inlined_filename);

  return result;
}

XePUtilsInlinedData *
a_Utils_data_has_binhex(const char *data, size_t size)
{
  const char *data_end = data + size;
  const char *begin_tag = "(This file must be converted with BinHex 4.0)";
  const char *begin;
  const char *begin_eol;
  const char *end;
  const char *end_eol;
  char *inlined_filename;
  XePUtilsInlinedData *result = NULL;
  int found = 0;

  if (data == NULL || size == 0)
    return NULL;

  /* We should loop until no "begin_tag" is found */

  begin = a_Utils_memmem(data, size, begin_tag, strlen(begin_tag));
  if (!begin)
    return NULL;

  /* NOTE: GNU libc claims that strchr(begin_tag, '\0') 
   * is faster than strlen 
   */
  begin_eol = begin + strlen(begin_tag);

  /* Fail if begin_tag is not on a separate line */

  if (!(begin == data || *(begin - 1) == '\n' || *(begin - 1) == '\r') ||
      !(begin_eol < data_end && (*begin_eol == '\n' || *begin_eol == '\r')))
    return NULL;

  /* There is no filename to look for */

  end = begin_eol;
  while (end < data_end && (*end == '\n' || *end == '\r'))
    {
      end++;
    }

  /* Fail is there is no data */

  if (end == data_end || *end != ':')
    return NULL;

  /* BixHex are lines of 64 characters. Only the last may have 65 where
   * the 65th character is the trailing ':'
   */
  end_eol = end + 1;
  while (end_eol < data_end && !found)
    {
      while (end_eol < data_end && 
	     *end_eol != '\r' &&
	     *end_eol != '\n' &&
	     *end_eol != ':')
	{
	  end_eol++;
	}
      if (end_eol == data_end)
	return NULL;

      /* Fail if intermediate lines are not 64 bytes long */

      if (*end_eol != ':' && (end_eol - end) != 64)
	return NULL;
      if (*end_eol == ':' && (end_eol - end) >= 1 && (end_eol - end) <= 65)
	found = 1;
      else
	{
	  while (end_eol < data_end && 
		 (*end_eol == '\r' || *end_eol == '\n'))
	    {
	      end_eol++;
	    } 
	  end = end_eol;
	}
    }

  if (found)
    {
      inlined_filename = strdup("BinHex 4.0 Encoded File");
      result = a_Utils_inlined_file_new(XEP_INLINED_DATA_BINHEX,
					begin,
					end_eol - begin + 1,
					inlined_filename, NULL);
      x_free(inlined_filename);
    }
  return result;
}

/*! Convert a message original date to a local time
 * \param date_time the message's original date field
 * \param local_tm the result. It is also used for intermediary computation
 * \return 0 on success, -1 on failure
 * Not reentrant because it uses localtime and not localtime_r
 */
int
a_Utils_orig_date_to_localtime(struct mailimf_date_time *date_time,
			       struct tm *local_tm)
{
  time_t a_time;
  time_t dtz;

  int abs_zone;
  struct tm *result;

  abs_zone = abs(date_time->zone);
  dtz = ((abs_zone % 100) + (abs_zone/100) * 60) * 60;
  if (date_time->zone < 0)
    {   
      dtz = -dtz;
    }

  local_tm->tm_sec   = date_time->sec;
  local_tm->tm_min   = date_time->min;
  local_tm->tm_hour  = date_time->hour;
  local_tm->tm_mday  = date_time->day;
  local_tm->tm_mon   = date_time->month - 1;
  /* Years with 2 digits are considered belonging to 20th century if
   * greater or equal than 72, 1972 being the birth of email
   * 0-71  -> 2000-2071
   * 72-1899 -> 1972-3799!
   * 1900-> correct year.
   * This is because the tm struct cannot handle years before 1900
   */
     
  local_tm->tm_year  = (date_time->year >= 1900 ? 
			date_time->year - 1900 :
			(date_time->year >= 72 ?
			 date_time->year :
			 date_time->year + 100));

  local_tm->tm_wday  = 0;
  local_tm->tm_yday  = 0;
  local_tm->tm_isdst = 0;

  a_time = mktime(local_tm);

  if (a_time != (time_t) -1)
    {
      a_time += -timezone - dtz;
      
      result = localtime(&a_time);
    }
  else
    {
      /* Consider incorrect dates to be on 01-01-1970 */
      a_time = 0; 
      result = gmtime(&a_time);
    }

  *local_tm = *result;

  return 0;
}

void *
a_Utils_memmem(const void *haystack, 
	       size_t haystack_len, 
	       const void *needle,
	       size_t needle_len)
{
#ifdef HAVE_MEMMEM_DISABLED
  return memmem(haystack, haystack_len, needle, needle_len);
#else
  const unsigned char *hs_p;
  const unsigned char *hs_oldp = (const unsigned char *) haystack;
  const unsigned char *n_p = (const unsigned char *) needle;
  const unsigned char *e_p = haystack + haystack_len;
  size_t rem_len = haystack_len;
  size_t i;

  while ((hs_oldp < e_p) && (hs_p = memchr(hs_oldp, *n_p, rem_len)))
    {
      rem_len -= hs_p - hs_oldp;
      if (rem_len < needle_len)
	return NULL;
      i = 1;
      for (i = 1; (i < needle_len) && (hs_p[i] == n_p[i]); i++)
	{
	  /* Empty on purpose: the search is done in the for loop */
	}

      if (i == needle_len)
	return (void *) hs_p;
      hs_oldp = hs_p + 1;
      rem_len--;
    }
  return NULL;
#endif
}


int
a_Utils_quote_text_for_reply(const char *text, 
			     size_t text_size,
			     const char *text_charset,
			     char **quoted_text)
{
  const char *text_p;
  const char *text_end = text + text_size;
  char *quoted_p;
  char *reply_prefix = "> ";
  int lines = 0;

  /* TODO: Depending on the charset (new lines?), insert quoting prefix 
   * For the moment, just use the original text
   */
  if (text == NULL)
    {
      *quoted_text = NULL;
      return 0;
    }

  /* TODO: use a robust algorithm for line detection.
   * This depends on the used character set
   * Unicode defines LSEP line separator (0x2028), PSEP (0x2029) paragraph
   * separator and NL (0x0085) Next Line
   */
  for (text_p = text; text_p < text_end; text_p++)
    {
      lines++;
      while (text_p < text_end && 
	     !(*text_p == '\r' && text_p + 1 < text_end && text_p[1] == '\n'))
	{
	  text_p++;
	}
      if (text_p + 1 < text_end && text_p[1] == '\n')
	{
	  text_p++;
	}
    }

  *quoted_text = x_alloc(text_size + lines*strlen(reply_prefix) + 1, char);
  quoted_p = *quoted_text;

  for (text_p = text; text_p < text_end; *text_p++, *quoted_p++)
    {
      memcpy(quoted_p, reply_prefix, strlen(reply_prefix));
      quoted_p += strlen(reply_prefix);
      while (text_p < text_end && 
	     !(*text_p == '\r' && text_p + 1 < text_end && text_p[1] == '\n'))
	{
	  *quoted_p = *text_p;
	  quoted_p++;
	  text_p++;
	}
      if (text_p + 1 < text_end && text_p[1] == '\n')
	{
	  *quoted_p = *text_p;
	  quoted_p++;
	  text_p++;
	  *quoted_p = *text_p;
	}
    }
  *quoted_p = 0x00;
  return 0;
}

clist *
a_Utils_parse_mailto(const char *mailto)
{
  clist *list = NULL;
  XePUtilsPair *header;

  /* TODO: implement real mailto parsing */

  if (mailto)
    {
      list = clist_new();
      header = x_alloc(1, XePUtilsPair);
      
      header->name = strdup("To: ");
      header->value = strdup(mailto);
      clist_append(list, header);
    }

  return list;  
}

int
a_Utils_exec_cmd(char *cmd, int fd_in, int fd_out, int fd_err, int *fd_mon,
		 pid_t *pid, int sync)
{
  int child_mon[2];
  int ret;

  pid_t child_pid, gdchild_pid;

  if (!cmd)
    {
      A_LOGGING_ERROR("No command given.");
      return -1;
    }

  if (sync && !fd_mon)
    {
      A_LOGGING_SYSTEM_ERROR("synchroneous command must be monitored.");
      return -1;
    }
  
  child_mon[0] = child_mon[1] = -1;

  if (sync && pipe(child_mon))
    {
      A_LOGGING_SYSTEM_ERROR("pipe: %s. Command '%s' aborted.", 
			     strerror(errno),
			     cmd);
      close(child_mon[0]);
      close(child_mon[1]);
      
      return -1;
    }

  child_pid = fork();

  if (child_pid != 0)
    {
      /* Parent process executes this block */
      if (child_pid < 0)
	{
	  /* Child forking did not succeed */
	  A_LOGGING_SYSTEM_ERROR("forking child: %s. Command '%s' aborted.", 
				 strerror(errno),
				 cmd);
	  if (sync)
	    {
	      close(child_mon[0]);
	      close(child_mon[1]);
	    }
	  ret = -1;
	}
      else
	{
	  /* Child forking succeded */
	  if (sync)
	    {
	      close(child_mon[1]);
	      *fd_mon = child_mon[0];
	      if (pid)
		*pid = child_pid;
	    }
	  else
	    waitpid(child_pid, NULL, 0);
	  ret = 0;
	}
    }
  else
    {
      /* Child process executes this block */

      int fd;
      int max_fd;
      int gdchild_status;

      if (setpgid(0, 0))
	A_LOGGING_SYSTEM_ERROR("setpgid: %s", strerror(errno));
      
      /* Close all file descriptors except                         */
      /* stdin, stdout, stderr, fd_in, fd_out, fd_err child_mon[1] */
      /* NB: if (sync), child_mon[1] == -1                         */
      max_fd = sysconf(_SC_OPEN_MAX);
      for (fd = 3; fd < max_fd; fd++)
	if (fd != child_mon[1] && fd != fd_in &&
	    fd != fd_out       && fd != fd_err)
	  close(fd);

      gdchild_pid = fork();

      if (gdchild_pid != 0)
	{
	  /* Child process executes this block */
	  if (gdchild_pid < 0)
	    {
	      /* Grand child forking error */
	      A_LOGGING_SYSTEM_ERROR("child process: forking grand child: %s."
				     "Command '%s' aborted.",
				     strerror(errno),
				     cmd);
	      if (sync)
		{
		  gdchild_status = 0; /* Irrelevant, no grand child forked */
		  write(child_mon[1], (char *) &gdchild_status, sizeof(int));
		  close(child_mon[1]);
		}
	      ret = 1;
	    }
	  else
	    {
	      /* Grand child forked succesfully */
	      if (sync)
		{
		  A_LOGGING_DEBUG("child process: waiting for grand child.");

		  waitpid(gdchild_pid, &gdchild_status, 0);

		  A_LOGGING_DEBUG("child process: grand child exited with "
				  "status %d", gdchild_status);

		  write(child_mon[1], (char *) &gdchild_status, sizeof(int));
		  close(child_mon[1]);
		}
	      else
		{
		  A_LOGGING_DEBUG("child process: grand child launched. "
				  "May it live happy.");
		}
	      ret = 0;
	    }
	  _exit(ret);
	}
      else
	{
	  /* Grand child executes this block */
	  char *cmdline[4];

	  close(child_mon[1]);

	  if (setpgid(0, getppid()))
	    A_LOGGING_SYSTEM_ERROR("grand child process: setpgid: %s.",
				   strerror(errno));
	  /* If fd_* are negative, duping is not done 
	     and the command inherits parent's standard files */
	  
	  dup2(fd_in,  STDIN_FILENO);
	  dup2(fd_out, STDOUT_FILENO);
	  dup2(fd_err, STDERR_FILENO);

	  cmdline[0] = "/bin/sh";
	  cmdline[1] = "-c";
	  cmdline[2] = cmd;
	  cmdline[3] = 0;

	  execv(cmdline[0], cmdline);
	  
	  A_LOGGING_SYSTEM_ERROR("grand child process: execv: %s. "
				 "Command '%s' aborted.", 
				 strerror(errno), 
				 cmd);
	  _exit(1);
	  /* End of grand child block */
	}
    }

  return ret;
}

char *
a_Utils_basename(const char *path)
{
  char *base_name = NULL;

  if (path == NULL || *path == 0x00)
    {
      base_name = strdup(".");
    }
  else if (strcmp(path, XETPAN_DIR_SEPARATOR_S) == 0)
    {
      base_name = strdup(XETPAN_DIR_SEPARATOR_S);
    }
  else
    {
      char *p;
      /* FIXME: this assumes that the charset allows the univoque
       * search of a character: this may not be always the case.
       */
      base_name = strdup(path);
      if (base_name)
	{
	  /* Get rid of an eventual trailing path separator */
	  if (base_name[strlen(base_name)] == XETPAN_DIR_SEPARATOR_C)
	    {
	      base_name[strlen(base_name)] = 0x00;
	    }

	  p = strrchr(base_name, XETPAN_DIR_SEPARATOR_C);
	  if (p)
	    {
	      int i;
	      for (i = 0; p[i]; i++)
		{
		  base_name[i] = p[i + 1];
		}
	    }

	}
    }
  if (base_name == NULL)
    {
      A_LOGGING_ERROR("strdup(): %s", strerror(errno));
    }
  return base_name;
}

char *
a_Utils_magic_get_bare_mime_type(const char *mime_type)
{
  char *bare_mime_type = NULL;
  char *p;
  int i = 1;

  if (mime_type == NULL)
    return NULL;

  /* Find the start of the subtype */
  p = strchr(mime_type, '/');

  if (p)
    {
      /* Find the end of the subtype */
      for (i = 1; 
	   (p[i] && 
	    !Utils_is_SPACE(p[i]) && 
	    !Utils_is_CTL(p[i]) && 
	    !Utils_is_tspecials(p[i]));
	   i++)
	{
	  /* Empty on purpose */
	}
      /* type+subtype length: take into account the text before the subtype */
      i += p - mime_type;
    }
  else
    {
      i = strlen(mime_type);
    }

  bare_mime_type = x_alloc(i + 1, char);
  
  if (bare_mime_type == NULL)
    {
      A_LOGGING_ERROR("x_alloc(): %s", strerror(errno));
      return NULL;
    }

  memcpy(bare_mime_type, mime_type, i);
  bare_mime_type[i] = 0x00;

  return bare_mime_type;
}

static int
Utils_is_WSP(unsigned char c)
{
  /* SPACE || HTAB */
  return ((c == 32) || (c == 9));
}

static int
Utils_is_NO_WS_CTL(unsigned char c)
{
  /* {1,...,31} \ {10, 13} U {127} */
  return ((c >= 1 && c <= 8) || (c == 11) || (c ==12) || 
	  (c >= 14 && c <= 31) || (c == 127));
}

static int
Utils_is_CTL(unsigned char c)
{
  /* rfc 2045 */
  return ((c <= 31) || (c == 127));
}

static int
Utils_is_SPACE(unsigned char c)
{
  return (c == ' ');
}

static int
Utils_is_ALPHA(unsigned char c)
{
  /* {65,...,90} U {97,..., 122} */
  return ((c >= 65 && c <= 90) || (c >= 97 && c <= 122));
}

static int
Utils_is_DIGIT(unsigned char c)
{
  /* {48,...,57} */
  return ((c >= 48 && c <= 57));
}

/* s[0] and s[1] must be referencable */
static int
Utils_is_quoted_pair(unsigned char *s)
{
  return (s[0] == '\\' && Utils_is_text(s[1]));
}

static int
Utils_is_text(unsigned char c)
{
  /* RFC 2822 */
  /* {1, 127} \ {10, 13} */
  return ((c >= 1 && c <= 9) || (c == 11) || (c ==12) || 
	  (c >= 14 && c <= 127));
}

static int
Utils_is_atext(unsigned char c)
{
  /* {48,...,57} U {65,...,90} U {97,..., 122} */
  /* 33 35-39 42 43 45 47 61 63 94 95 96 123 124  126  */
  /*  0-9 A-Z a-z */
  /*  ! # $ % & ' * + - / = ? ^ _ ` { } ~ */

  return ((c == 33) || (c >= 35 && c <= 39) || (c == 42) || (c == 43) ||
	  (c == 45) || (c >= 47 && c <= 57) || (c == 61) || (c == 63) ||
	  (c >= 65 && c <= 90) || (c >= 94 && c <= 96) || (c == 123) ||
	  (c == 124) || (c == 126));
}

static int
Utils_is_ctext(unsigned char c)
{
  /* {1,...,126} \ {10, 13, 32, 40, 41, 92} */
  return ((Utils_is_NO_WS_CTL(c)) || (c >= 33 && c <= 39) ||
	  (c >= 42 && c <= 91) || (c >= 93 && c <= 126));
}

static int
Utils_is_qtext(unsigned char c)
{
  /* {1,...,127} \ {10, 13, 32, 34, 92} 
   * All US-ASCII except NUL CR LF SPACE " \ 
   */

  return ((Utils_is_NO_WS_CTL(c)) || (c == 33) || (c >= 35 && c <= 91) ||
	  (c >= 93 && c <= 126));
	  
}

static int
Utils_is_qcontent(unsigned char *s)
{
  return (Utils_is_qtext(s[0]) || Utils_is_quoted_pair(s));
}

static int
Utils_is_utext(unsigned char c)
{
  return ((Utils_is_NO_WS_CTL(c)) || (c >= 33 && c <= 126));
}

static int
Utils_is_specials(unsigned char c)
{
  return ((c == '(') || (c == ')') || (c == '<') || (c == '>') ||
	  (c == '[') || (c == ']') || (c == ':') || (c == ';') ||
	  (c == '@') || (c == '\\') || (c == ',') || (c == '.') ||
	  (c == '"'));
}

static int
Utils_is_especials(unsigned char c)
{
  return ((c == '(') || (c == ')') || (c == '<') || (c == '>') || (c == '@') ||
	  (c == ',') || (c == ';') || (c == ':') || (c == '"') || (c == '/') ||
	  (c == '[') || (c == ']') || (c == '?') || (c == '.') || (c == '='));
}
static int
Utils_is_tspecials(unsigned char c)
{
  return (c == '(' || c == ')' || c == '<' || c == '>' ||
	  c == '@' || c == ',' || c == ':' || c == '\\' ||
	  c == '"' || c == '/' || c == '[' || c == ']' ||
	  c == '?' || c == '=');
}

static int
Utils_is_rfc822_quoted_pair(unsigned char *s)
{
  return (s[0] == '\\' && Utils_is_rfc822_text(s));
}

/* s[0] and s[1] must be referencable */
static int
Utils_is_rfc822_text(unsigned char *s)
{
  /* RFC822 */
  /* [0, 127] / !(s[0] == 13 (CR) && s[1] == 10 (LF)) */
  
  return ((s[0] <= 127) && !((s[0] == 13) && (s[1] == 10)));
}

void
a_Utils_to_b_encoding(const char *s, size_t *is, char *t, size_t *it)
{
  const char base64[] = 
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    
  unsigned char a, b, c;
  s += *is;
  t += *it;

  a = s[0];
  if (s[1])
    {
      b = s[1];
      if (s[2])
	{
	  c = s[2];
	  *is += 3;
	  t[0] = base64[a >> 2];
	  t[1] = base64[((a & 0x03) << 4) | (b >> 4)];
	  t[2] = base64[((b & 0x0F) << 2) | (c >> 6)];
	  t[3] = base64[(c & 0x3f)];
	  *it += 4;
	  return;
	}
      else
	{
	  *is += 2;
	  t[0] = base64[a >> 2];
	  t[1] = base64[((a & 0x03) << 4) | (b >> 4)];
	  t[2] = base64[((b & 0x0F) << 2)];
	  t[3] = '=';
	  *it += 4;
	  return;
	}
    }
  else  
    {
      *is += 1;
      t[0] = base64[a >> 2];
      t[1] = base64[((a & 0x03)<< 4)];
      t[2] = '=';
      t[3] = '=';
      *it += 4;
      return;
    }
}

static int
Utils_is_to_be_q_encoded(char c, int is_phrase, int is_comment)
{
return ((c >= 'a' && c <= 'z') || 
	(c >= 'A' && c <= 'Z') || 
	(c >= '0' && c <= '9') ||
	(c == '*') || (c == '+') ||
	(c == '-') || (c == '/') ||
	(!is_phrase && (c == '\'' || (c == '.') || (c == ',') || (c == ':'))) ||
	(!is_phrase && !is_comment && ((c == '(') || c == ')')));
}

void
a_Utils_to_q_encoding(const char *s, size_t *is, char *t, size_t *it, 
		      int is_phrase,
		      int is_comment)
{
  char hex[] = "0123456789ABCDEF";
  unsigned char c = s[*is];
  t += *it;
  if (Utils_is_to_be_q_encoded(c, is_phrase, is_comment))
    {
      t[0] = c; *it += 1; *is += 1;
      return;
    }
  t[0] = '='; t[1] = hex[(c >> 4)]; t[2] = hex[c & 0x0F]; 
  *it += 3;
  *is += 1;
}

static XePUtilsEncodingType
Utils_select_b_q_encoding(const char *text, const char *charset, 
			  int is_phrase, int is_comment)
{
  const char *p;
  size_t qp = 0;
  for (p = text; *p; p++)
    {
      if (Utils_is_to_be_q_encoded(*p, is_phrase, is_comment))
	{
	  qp++;
	}
    }
  /* Approximate formula to tell which encoding is shorter
   * The exact equation is:
   * (p - text) - 6*qp for text length multiple of 3
   * (p - text) + 4 - 6*qp when text length modulo 3 == 2
   * (p - text) + 8 - 6*qp when text length modulo 3 == 1
   * B encoding being chosen when the latter expression is negative
   * and Q when positive
   */
  if ((p - text +  4) < 6 * qp) 
    return XEP_B_ENCODING;
  else
    return XEP_Q_ENCODING;  
}
      
char *
a_Utils_encode_8bit_header(const char *text, 
			   const char *charset,
			   XePUtilsEncodingType encoding, 
			   int is_phrase, 
			   int is_comment)
{
  size_t is = 0;
  size_t ir = 0;
  size_t len_text;
  size_t len_result;
  char *result;
  size_t len_base = strlen("=??Q?=");
  if (text == NULL)
    return NULL;

  if (*text != 0x00 && charset != NULL)
    {
      len_text = strlen(text);
      
      if (encoding == XEP_AUTO_ENCODING)
	{
	  encoding = Utils_select_b_q_encoding(text, charset,
					       is_phrase, is_comment);
	}
      /* Guess the maximal size of the encoded data */

      len_result = (encoding == 1 ? len_text * 3 : ((len_text + 3) * 4) / 3);

      /* Count the charset part and the base "=??Q?=" */

      len_result += strlen(charset) + len_base;

      result = x_alloc(len_result + 1, char);

      if (result)
	{     
	  sprintf(result, "=?%s?%c?", charset, (encoding == XEP_Q_ENCODING ? 'Q' : 'B'));
	  ir = strlen(result);
	  if (encoding == XEP_Q_ENCODING)
	    {
	      while (is < len_text && ir < len_result)
		{
		  a_Utils_to_q_encoding(text, &is, result, &ir, 
					is_comment, is_comment);
		}
	    }
	  else /* B Encoding */
	    {
	      while (is < len_text && ir < len_result)
		{
		  a_Utils_to_b_encoding(text, &is, result, &ir);
		}
	    }
	  result[ir] = 0x00;
	  if (is + 3 >= len_result)
	    {
	      char *tmp;
	      printf("Should not...\n");
	      tmp = realloc(result, len_result + 3);
	      if (tmp)
		{
		  result = tmp;
		}
	      else
		{
		  A_LOGGING_ERROR("realloc(): %s", strerror(errno));
		  x_free(result);
		  return NULL;
		}
	    }
	  strcat(result, "?=");
	}
      else
	{
	  A_LOGGING_ERROR("x_alloc(): %s", strerror(errno));
	}
    }
  else
    {
      result = strdup(text);
      if (result == NULL)
	{
	  A_LOGGING_ERROR("strdup(): %s", strerror(errno));
	}
    }
  return result;
}

char *
a_Utils_encode_8bit_subject(const char *subject, 
			    const char *charset, 
			    XePUtilsEncodingType encoding)
{
  char *result;
  
  if (subject == NULL)
    return NULL;

  if (a_Utils_has_8bit(subject, 0))
    {
      result = a_Utils_encode_8bit_header(subject, charset, encoding, 0, 0);
    }
  else
    {
      result = strdup(subject);
      if (result == NULL)
	{
	  A_LOGGING_ERROR("strdup(): %s", strerror(errno));
	}
    }
  return result;
}

int
a_Utils_has_8bit(const char *text, size_t size)
  {
  const unsigned char *p;

  if (!size)
    {
      /* Loop until NUL byte or a 8bit byte */
      for (p = text; p[0] && !(p[0] & 0x80); p++);
    }
  else
    {
      size_t i;
      /* Loop until NUL byte or a 8bit byte */
      for (p = text, i = 0; p[0] && !(p[0] & 0x80) && i < size; p++, i++);
    }

  /* Loop did not end at the NUL byte ? Then there were 8bit byte */
  return ((p[0] & 0x80) != 0);
}

#define SCAN_FUNC_NB 4
clist *
a_Utils_parse_body_text(const char *text, size_t size)
{
  const char *data_p;
  clist *result;
  size_t remaining_len;
  XePUtilsTextScanFunc scan_func[SCAN_FUNC_NB] =
    {a_Utils_data_has_uuencoded,
     a_Utils_data_has_yencoded,
     a_Utils_data_has_yencoded2,
     a_Utils_data_has_binhex
     /* a_Utils_data_has_clearsigned, */
     /* a_Utils_data_has_ascii_armored */
    };
  XePUtilsInlinedData *inlined_data[SCAN_FUNC_NB];
  XePUtilsInlinedData *first_data = NULL;
  XePUtilsInlinedData *prev_data = NULL;
  XePUtilsInlinedData *text_data = NULL;
  int i_found[SCAN_FUNC_NB];
  int i;


  result = clist_new();
  if (result == NULL)
    {
      A_LOGGING_ERROR("clist_new() failed: %s", strerror(errno));
      return NULL;
    }

  for (i = 0; i < SCAN_FUNC_NB; i++)
    {
      i_found[i] = 1;
      inlined_data[i] = NULL;
    }

  remaining_len = size;
  data_p = text;

  while (remaining_len)
    {
      /* Check for uuencoded, yEnc or yEnc2 or BinHex 4.0 data.
       * Check again only if something was found the last time we checked.
       */
      prev_data = first_data;
      first_data = NULL;
      for (i = 0; i < SCAN_FUNC_NB; i++)
	{
	  if (i_found[i])
	    {
	      /* We do not need to scan again for a certain type of inlined
	       * data if this very type was found in the previous iteration
	       * and if it was not processed because another type was found before it.
	       * Thus, we scan only when it is the first ever scan of this type
	       * or when we already processed this very inlined_data at 
	       * the previous iteration.
	       */
	      if ((inlined_data[i] == NULL) ||
		  (prev_data && inlined_data[i]->any.start == prev_data->any.start))
		{
		  inlined_data[i] = scan_func[i](data_p, remaining_len);
		}

	      if (inlined_data[i])
		{
		  if ((!first_data) ||
		      (inlined_data[i]->any.start <= first_data->any.start))
		    {
		      first_data = inlined_data[i];
		    }
		}
	      else
		{
		  i_found[i] = 0;
		}
	    }
	}

      if (first_data)
	{
	  /* First the text placed before the encoded file */
	  if (data_p < first_data->any.start)
	    {
	      text_data = a_Utils_inlined_text_new(XEP_INLINED_DATA_TEXT,
						   data_p,
						   first_data->any.start - data_p,
						   NULL, 0,
						   0, 0, NULL);
	      if (text_data)
		{
		  clist_append(result, text_data);
		}
	    }
	  clist_append(result, first_data);

	  data_p = first_data->any.start + first_data->any.size;
	  remaining_len = text + size - data_p;
	}
      else
	{
	  /* Nothing found: display all remaining text */
	  text_data = a_Utils_inlined_text_new(XEP_INLINED_DATA_TEXT,
					       data_p,
					       remaining_len,
					       NULL, 0,
					       0, 0, NULL);
	  remaining_len = 0;
	  if (text_data)
	    {
	      clist_append(result, text_data);
	    }
	}
    }

  if (clist_count(result) == 0)
    {
      clist_free(result);
      result = NULL;
    }

  return result;
}
#undef SCAN_FUNC_NB

const char *
a_Utils_inlined_data_type_as_string(XePUtilsInlinedDataType type)
{
  return (type == XEP_INLINED_DATA_UUENCODED ? 
	  _("uuencoding") :
	  type == XEP_INLINED_DATA_YENCODED ?
	  _("yEnc") :
	  type == XEP_INLINED_DATA_YENCODED2 ?
	  _("yEnc2") :
	  type == XEP_INLINED_DATA_BINHEX ?
	  _("BinHex 4.0") :
	  type == XEP_INLINED_DATA_TEXT ?
	  _("Text") :
	  type == XEP_INLINED_DATA_CLEARSIGNED_TEXT ?
	  _("Clear-signed text") :
	  type == XEP_INLINED_DATA_ASCII_ARMORED_TEXT ?
	  _("ASCII armored encrypted text") :
	  _("Unknown"));
}

int
a_Utils_text_new_copy_with_rfc2822_lines(const char *text, size_t size,
					 char **rfc2822_text,
					 size_t *rfc2822_size)
{
  MMAPString * mmapstr;
  size_t len;
  const char *sp = text;
  const char *sp_end = text +  size;

  mmapstr = mmap_string_sized_new(size);
  if (mmapstr == NULL)
    goto err;

  len = 0;
  for (; sp < sp_end - 1; sp++) {
    if (*sp != '\r' && *sp != '\n') {
      if (len < SMTP_MAX_LINE_LEN) { 
	/* most taken path:  (not \r\n) and (len < SMTP_MAX_LINE_LEN) */
	len++;
	continue;
      } 
      /* line reached SMTP_MAX_LINE_LEN and current character is not a 
       * line break */
      if (!mmap_string_append_len(mmapstr, sp - len, len) ||
	  !mmap_string_append(mmapstr, "\r\n"))
	goto free;
      len = 1; /* current character must be included in next stint */
      continue;
    }
    if (*sp != '\r' || sp[1] != '\n') {
      /* (not \r\n) but a kind of line break */
      if (!mmap_string_append_len(mmapstr, sp - len, len) ||
	  !mmap_string_append(mmapstr, "\r\n"))  /* canonical line break */
	goto free;
      len = 0; /* current character is a line break, do not include it in
		* next stint */
      continue;
    }
    /* \r\n */
    if (!mmap_string_append_len(mmapstr, sp - len, len + 2))
      goto free; /* \r\n */
    len = 0;
    sp++;
  }

  if (sp < sp_end) { /* last two bytes were not \r\n */
    if (*sp != '\r' && *sp != '\n') {
      if (len == SMTP_MAX_LINE_LEN) {
	if (!mmap_string_append_len(mmapstr, sp - len, len) ||
	    !mmap_string_append(mmapstr, "\r\n") ||
	    !mmap_string_append_len(mmapstr, sp, 1))
	  goto free;
      } else {
	if (!mmap_string_append_len(mmapstr, sp - len, len + 1))
	  goto free;
      }
    } else { /* \r or \n */
      if (!mmap_string_append_len(mmapstr, sp - len, len) ||
	  !mmap_string_append(mmapstr, "\r\n"))
	goto free;
    }
  }

  if (mmap_string_ref(mmapstr) < 0)
    goto free;

  *rfc2822_text = mmapstr->str;
  *rfc2822_size  = mmapstr->len;

  return 0;

 free:
  mmap_string_free(mmapstr);
 err:
  return -1;
}
